/*
 * i2c Support for Atmel's Two-Wire Interface (TWI)
 *
 * Based on the work of Copyright (C) 2004 Rick Bronson
 * Converted to 2.6 by Andrew Victor <andrew at sanpeople.com>
 * Ported to AVR32 and heavily modified by Espen Krangnes
 * <ekrangnes at atmel.com>
 *
 * Copyright (C) 2006 Atmel Corporation
 *
 * Borrowed heavily from the original work by:
 * Copyright (C) 2000 Philip Edelbrock <phil at stimpy.netroedge.com>
 *
 * Partialy rewriten by Karel Hojdar <cmkaho at seznam.cz>
 * bugs removed, interrupt routine markedly rewritten
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


#include <linux/err.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/completion.h>
#include <asm/io.h>
#include <linux/time.h>
#include "atmeltwi.h"

static unsigned int baudrate = CONFIG_I2C_ATMELTWI_BAUDRATE;
module_param(baudrate, uint, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(baudrate, "The TWI baudrate");


struct atmel_twi {
	void __iomem *regs;
	struct i2c_adapter adapter;
	struct clk *pclk;
	spinlock_t lock;
	struct completion comp;
	u32 intmask;
	u8 *buf;
	u8 len;
	u8 acks_left;
	u8 nack;
	unsigned int irq;

};
#define to_atmel_twi(adap) container_of(adap, struct atmel_twi, adapter)

/*
 * Initialize the TWI hardware registers.
 */
static int __devinit twi_hwinit(struct atmel_twi *twi)
{
	unsigned long cdiv, ckdiv=0;

	twi_writel(twi, IDR, ~0UL);
	twi_writel(twi, CR, TWI_BIT(SWRST));	/*Reset peripheral*/
	twi_readl(twi, SR);

	cdiv = (clk_get_rate(twi->pclk) / (2 * baudrate)) - 4;

	while (cdiv > 255) {
		ckdiv++;
		cdiv = cdiv >> 1;
	}

	if (ckdiv > 7)
		return -EINVAL;
	else
		twi_writel(twi, CWGR, (TWI_BF(CKDIV, ckdiv)
			       | TWI_BF(CHDIV, cdiv)
			       | TWI_BF(CLDIV, cdiv)));
	return 0;
}

/*
 * Waits for the i2c status register to set the specified bitmask
 * Returns 0 if timed out (~100ms).
 */
static short twi_complete(struct atmel_twi *twi, u32 mask)
{
	int timeout = msecs_to_jiffies(100);

	twi->intmask = mask;
	init_completion(&twi->comp);

	twi_writel(twi, IER, mask);

	if (!wait_for_completion_timeout(&twi->comp, timeout)) {
		/* RESET TWI interface */
		twi_writel(twi, CR, TWI_BIT(SWRST));

		/* Reinitialize TWI */
		twi_hwinit(twi);

		return -ETIMEDOUT;
	}
	return 0;
}

/*
 * Generic i2c master transfer entrypoint.
 */
static int twi_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
	struct atmel_twi *twi = to_atmel_twi(adap);
	struct i2c_msg *pmsg;
	int i;

	/* get first message */
	pmsg = msgs;

	dev_dbg(&adap->dev, "twi_xfer: processing %d messages:\n", num);

	twi->nack = 0;
	for (i = 0; i < num; i++, pmsg++) {
		twi->len = pmsg->len;
		twi->buf = pmsg->buf;
		twi->acks_left = pmsg->len;
		twi_writel(twi, MMR, TWI_BF(DADR, pmsg->addr) |
			(pmsg->flags & I2C_M_RD ? TWI_BIT(MREAD) : 0));
		twi_writel(twi, IADR, TWI_BF(IADR, pmsg->addr));

		dev_dbg(&adap->dev,"#%d: internal addr %d %s byte%s %s 0x%02x\n",
			i,pmsg->len, pmsg->flags & I2C_M_RD ? "reading" : "writing",
			pmsg->len > 1 ? "s" : "",
			pmsg->flags & I2C_M_RD ? "from" : "to", pmsg->addr);

		/* enable */
		twi_writel(twi, CR, TWI_BIT(MSEN));

		if (pmsg->flags & I2C_M_RD) {
			if (twi->len == 1)
				twi_writel(twi, CR,
					TWI_BIT(START) | TWI_BIT(STOP));
			else
				twi_writel(twi, CR, TWI_BIT(START));

			if (twi_complete(twi, TWI_BIT(RXRDY)) == -ETIMEDOUT) {
				dev_dbg(&adap->dev, "RXRDY timeout. Stopped with %d bytes left\n",
					twi->acks_left);
				return -ETIMEDOUT;
			}
		} else {
			twi_writel(twi, THR, twi->buf[0]);
			if (twi_complete(twi, TWI_BIT(TXRDY)) == -ETIMEDOUT) {
				dev_dbg(&adap->dev, "TXRDY timeout. Stopped with %d bytes left\n",
				twi->acks_left);
				return -ETIMEDOUT;
			}

			if (twi->nack)
			    return -ENODEV;
		}

		/* Disable TWI interface */
		twi_writel(twi, CR, TWI_BIT(MSDIS));

	} /* end cur msg */

	return i;
}


static irqreturn_t twi_interrupt(int irq, void *dev_id)
{
	struct atmel_twi *twi = dev_id;
	int status = twi_readl(twi, SR);

	/* Save state for later debug prints */
	int old_mask = twi->intmask;
	int old_status = status;

	if (twi->intmask & status) {
		if (status & TWI_BIT(NACK))
			goto nack;

		status &= twi->intmask;

		if (status & TWI_BIT(TXCOMP))
		    goto complete;

		else if (status & TWI_BIT(RXRDY)) {
			if ( twi->acks_left > 0 ) {
			    twi->buf[twi->len - twi->acks_left] =
				twi_readl(twi, RHR);
			    twi->acks_left--;
			}
			if ( twi->acks_left == 1 )
				twi_writel(twi, CR, TWI_BIT(STOP));

			if (twi->acks_left == 0 ) {
				twi->intmask = TWI_BIT(TXCOMP);
				twi_writel(twi, IER, TWI_BIT(TXCOMP));
			}
		} else if (status & TWI_BIT(TXRDY)) {
			twi->acks_left--;
			if ( twi->acks_left == 0 ) {
				twi->intmask = TWI_BIT(TXCOMP);
				twi_writel(twi, IER, TWI_BIT(TXCOMP));
			} else if (twi->acks_left > 0)
				twi_writel(twi, THR, twi->buf[twi->len - twi->acks_left]);
		}
	}

	dev_dbg(&twi->adapter.dev,
		"TWI ISR, SR 0x%08X, intmask 0x%08X, acks_left %i.\n",
		old_status, old_mask, twi->acks_left);

	return IRQ_HANDLED;

nack:
	dev_dbg(&twi->adapter.dev, "NACK received!\n");
	twi->nack = 1;

complete:
	twi_writel(twi, IDR, ~0UL);
	complete(&twi->comp);

	dev_dbg(&twi->adapter.dev,
		"TWI ISR, SR 0x%08X, intmask 0x%08X, \
		acks_left %i - completed.\n",
		old_status, old_mask, twi->acks_left);

	return IRQ_HANDLED;
}


/*
 * Return list of supported functionality.
 */
static u32 twi_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

/* For now, we only handle combined mode (smbus) */
static struct i2c_algorithm twi_algorithm = {
	.master_xfer	= twi_xfer,
	.functionality	= twi_func,
};

/*
 * Main initialization routine.
 */
static int __devinit twi_probe(struct platform_device *pdev)
{
	struct atmel_twi *twi;
	struct resource *regs;
	struct clk *pclk;
	struct i2c_adapter *adapter;
	int rc, irq;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	pclk = clk_get(&pdev->dev, "pclk");
	if (IS_ERR(pclk))
		return PTR_ERR(pclk);
	clk_enable(pclk);

	rc = -ENOMEM;
	twi = kzalloc(sizeof(struct atmel_twi), GFP_KERNEL);
	if (!twi) {
		dev_err(&pdev->dev, "can't allocate interface!\n");
		goto err_alloc_twi;
	}

	twi->pclk = pclk;
	twi->regs = ioremap(regs->start, regs->end - regs->start + 1);
	if (!twi->regs)
		goto err_ioremap;

	irq = platform_get_irq(pdev,0);
	rc = request_irq(irq, twi_interrupt, 0, "twi", twi);
	if (rc) {
		dev_err(&pdev->dev, "can't bind irq!\n");
		goto err_irq;
	}
	twi->irq = irq;

	rc = twi_hwinit(twi);
	if (rc) {
		dev_err(&pdev->dev, "Unable to set baudrate\n");
		goto err_hw_init;
	}

	adapter = &twi->adapter;
	sprintf(adapter->name, "TWI");
	adapter->algo = &twi_algorithm;
	adapter->class = I2C_CLASS_HWMON;
	adapter->dev.parent = &pdev->dev;

	platform_set_drvdata(pdev, twi);

	rc = i2c_add_adapter(adapter);
	if (rc) {
		dev_err(&pdev->dev, "Adapter %s registration failed\n",
			adapter->name);
		goto err_register;
	}

	dev_info(&pdev->dev, "Atmel TWI i2c bus device (baudrate %dk) at 0x%08lx.\n",
		 baudrate/1000, (unsigned long)regs->start);

	return 0;


err_register:
	platform_set_drvdata(pdev, NULL);

err_hw_init:
	free_irq(irq, twi);

err_irq:
	iounmap(twi->regs);

err_ioremap:
	kfree(twi);

err_alloc_twi:
	clk_disable(pclk);
	clk_put(pclk);

	return rc;
}

static int __devexit twi_remove(struct platform_device *pdev)
{
	struct atmel_twi *twi = platform_get_drvdata(pdev);
	int res;

	platform_set_drvdata(pdev, NULL);
	res = i2c_del_adapter(&twi->adapter);
	twi_writel(twi, CR, TWI_BIT(MSDIS));
	iounmap(twi->regs);
	clk_disable(twi->pclk);
	clk_put(twi->pclk);
	free_irq(twi->irq, twi);
	kfree(twi);

	return res;
}

static struct platform_driver twi_driver = {
	.probe		= twi_probe,
	.remove		= __devexit_p(twi_remove),
	.driver		= {
		.name	= "atmel_twi",
		.owner	= THIS_MODULE,
	},
};

static int __init atmel_twi_init(void)
{
	return platform_driver_register(&twi_driver);
}

static void __exit atmel_twi_exit(void)
{
	platform_driver_unregister(&twi_driver);
}

module_init(atmel_twi_init);
module_exit(atmel_twi_exit);

MODULE_AUTHOR("Espen Krangnes");
MODULE_DESCRIPTION("I2C driver for Atmel TWI");
MODULE_LICENSE("GPL");
